<?php declare(strict_types=1);

namespace Drupal\group_clone\Form;

use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Config\Config;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Form\FormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Routing\RouteBuilderInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

final class GroupCloneSettingsForm extends FormBase {

  use GroupCloneFormTrait;

  private EntityTypeManagerInterface $entityTypeManager;
  private RouteBuilderInterface $routeBuilder;

  /**
   * Constructs a new GroupCloneSettingsForm object.
   */
  public function __construct(
    EntityTypeManagerInterface $entity_type_manager,
    RouteBuilderInterface $route_builder
  ) {
    $this->entityTypeManager = $entity_type_manager;
    $this->routeBuilder = $route_builder;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container): self {
    return new static(
      $container->get('entity_type.manager'),
      $container->get('router.builder')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId(): string {
    return 'group_clone_settings';
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state): array {
    $group_content_type_storage = $this->entityTypeManager->getStorage('group_content_type');
    $group_types = $this->entityTypeManager->getStorage('group_type')->loadMultiple();

    foreach ($group_types as $group_type) {
      $id = $group_type->id();
      $config_entity = $this->getConfigEntity($id);

      $form[$id] = [
        '#tree' => TRUE,
        '#type' => 'details',
        '#open' => $config_entity->get('status'),
        '#title' => $group_type->label(),
      ];
      $form[$id]['status'] = [
        '#type' => 'checkbox',
        '#title' => $this->t('Enabled'),
        '#default_value' => $config_entity->get('status'),
      ];
      $form[$id]['content_types'] = $this->getGroupTypeSettingsElement($group_content_type_storage, $group_type, $config_entity);
    }

    $form['actions'] = ['#type' => 'actions'];
    $form['actions']['submit'] = [
      '#type' => 'submit',
      '#value' => $this->t('Save'),
    ];

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state): void {
    $form_state->cleanValues();
    foreach ($form_state->getValues() as $id => $settings_array) {
      foreach ($settings_array['content_types'] as $type_id => $value) {
        if ($value === 'disabled') {
          unset($settings_array['content_types'][$type_id]);
        }
        else {
          $settings_array['content_types'][$type_id] = $value;
        }
      }

      $config_entity = $this->getConfigEntity($id);
      foreach ($settings_array as $property => $value) {
        $config_entity->set($property, $value);
      }

      if (!$config_entity->empty()) {
        $config_entity->save();
      }
      elseif (!$config_entity->isNew()) {
        $config_entity->delete();
      }
    }

    // Rebuild routes as the one defined by this module depends on those
    // config entities.
    $this->routeBuilder->rebuild();

    $this->messenger()->addStatus($this->t('Group clone setting saved.'));
  }

}
